import click
import os
from accorder.utils.configurations import run_configuration
from accorder.utils.uploads import run_upload
from accorder.utils.submits import run_submit
from accorder.utils.builds import run_build
from collections import OrderedDict

APP_NAME = "accorder"

# ---------------------------------------
# readline path completion by johann woelper
# from: https://schdbr.de/python-readline-path-completion/


try:
    import readline
except ImportError:
    try:
        import pyreadline as readline
    except ImportError:
        readline = None

def list_folder(path):
    if path.startswith(os.path.sep):
        basedir = os.path.dirname(path)
        contents = os.listdir(basedir)
        contents = [os.path.join(basedir, d) for d in contents]
        contents = [d + os.path.sep for d in contents if os.path.isdir(d)]
    else:
        contents = os.listdir(os.curdir)
        contents = [d + os.path.sep for d in contents if os.path.isdir(d)]
    return contents


def completer(text, state):
    options = [x for x in list_folder(text) if x.startswith(text)]
    return options[state]


readline.set_completer(completer)

if readline.__doc__ and "libedit" in readline.__doc__:
    readline.parse_and_bind("bind -e")
    readline.parse_and_bind("bind ^I rl_complete")
else:
    readline.parse_and_bind("tab: complete")
    readline.set_completer_delims(" \t\n`~!@#$%^&*()-=+[{]}\\|;:'\",<>?")
# end of readline


# ---------------------------------------
# click stuff


class NaturalOrderGroup(click.Group):
    def __init__(self, name=None, commands=None, **attrs):
        if commands is None:
            commands = OrderedDict()
        elif not isinstance(commands, OrderedDict):
            commands = OrderedDict(commands)
        click.Group.__init__(self, name=name, commands=commands, **attrs)

    def list_commands(self, ctx):
        return self.commands.keys()


def get_version():
    import pkg_resources

    return pkg_resources.get_distribution("accorder").version


@click.group(cls=NaturalOrderGroup)
@click.version_option(version=get_version())
def cli():
    """Accorder takes care of various tasks which Memory of the World amateur librarians do in order to maintain their shared catalogs online.\n
    It builds searchable, standalone, portable webapp which one could then just copy to USB disk and open BROWSE_LIBRARY.html in her web browser.\n
    It uploads all of the books and metadata from local Calibre's library (together with portable webapp) to the server.\n
    It helps a librarian to maintain and share her catalog at https://library.memoryoftheworld.org together with other amateur librarians.\n
    It does all of above in one go by typing: accorder release PROFILE.\n
    The configuration file will keep information about one or more PROFILE. Under every PROFILE's configuration section there will be information about the directory path of local Calibre's library, librarian's name, credentials needed to upload the files to the destination server etc.
    """
    pass


@cli.command(
    short_help="Manage configuration. For more help, type: accorder configuration --help"
)
@click.argument("configuration_id", required=False)
@click.option(
    "--show", "-s", help="Show summary from the configuration file.", is_flag=True
)
@click.option(
    "--replace",
    "-r",
    help="Replace current configuration with the new one set by CONFIGURATION_ID. This flag requires CONFIGURATION_ID to be provided. (e.g: `accorder configuration --replace CONFIGURATION_ID``)",
    is_flag=True,
)
def configuration(configuration_id, show, replace):
    run_configuration(APP_NAME, configuration_id, show, replace)


@cli.command(
    short_help="Build standalone, portable webapp against the local Calibre library. For more help, type: accorder build --help"
)
@click.argument("profile")
@click.option(
    "--directory",
    "-d",
    help="Add/change PROFILE's directory path (of Calibre library) for the standalone, portable webapp to be built including all the metadata needed. Once provided, the directory path is automatically saved into configuration file for the future use.",
    type=click.Path(exists=True, resolve_path=True, file_okay=False),
)
@click.option(
    "--librarian",
    "-l",
    help="Add/change librarian's name (max. 42 chars) for the given PROFILE. Once provided, librarian's name is automatically saved into configuration file for the future use.",
)
@click.option(
    "--jsonpath",
    "-j",
    help="Built/render all metadata as JSON ino the file for which the path is provided.",
    type=click.Path(file_okay=True),
)
@click.option(
    "--with-bibtex",
    "-b",
    help="Add books from Zotero's bibtex export. It only adds entries with files/attachments with extensions: pdf, epub, mobi, azw, azw3, cbz, chm, djvu, doc, docx, fb2, htmlz, mbp, md, txt, zip)",
    type=click.Path(file_okay=True),
)
def build(profile, directory, librarian, jsonpath, with_bibtex):
    """Build searchable, standalone, portable webapp against the local Calibre library including all the metadata needed. It creates BROWSE_LIBRARY.html in root directory of the local Calibre library. For search (authors, titles, tags...) it uses rendered metadata from static/data.js file.\n
    If the local Calibre's library directory path is not already provided in the configuration file `accorder` will ask you to provide it (TAB completion should work).
    The same directory path could be anytime changed by using the flag (-d, --directory). For example:\n
    accorder build -d /your/path/to/calibre/library/ PROFILE\n
    Every time the directory path is provided it is saved in configuration file for the future use (therefore: `accorder build PROFILE` should be enough for the successful build).
"""
    run_build(APP_NAME, profile, directory, librarian, jsonpath, with_bibtex)


@cli.command(
    short_help="Upload the Calibre collection to the Memory of the World Library. For more help, type: accorder upload --help"
)
@click.argument("profile")
@click.option(
    "--directory",
    "-d",
    help="Add/change PROFILE's directory path to be uploaded to the server. The directory path is automatically saved into configuration file for the future use.",
    type=click.Path(exists=True, resolve_path=True, file_okay=False),
)
@click.option(
    "--delete-residue",
    help="Delete any remote directories/files which are not present locally anymore.",
    is_flag=True,
)
def upload(profile, directory, delete_residue):
    """Upload the Calibre collection to the Memory of the World server. Rsync (or rclone) will only upload local files which are missing at (or different from) the remote destination at the Memory of the World server.\n
    If the local Calibre's library directory path is not already provided in the configuration file `accorder` will ask you to provide it (TAB completion should work).
    The same directory path could be anytime changed by using the flag (-d, --directory). For example:\n
    accorder upload -d /your/path/to/calibre/library/ PROFILE\n
    Every time the directory path is provided it is saved in configuration file for the future use (therefore: `accorder upload PROFILE` should be enough for the successful upload).
    """
    run_upload(APP_NAME, profile, directory, delete_residue)


@cli.command(
    short_help="Maintain the shared Calibre's collection at the Memory of the World Library. For more help: accorder submit --help"
)
@click.argument("profile")
@click.option(
    "--librarian",
    "-l",
    help="Add/change librarian's name (max. 42 chars) for the given PROFILE. Once provided, librarian's name is automatically saved into configuration file for the future use.",
)
@click.option(
    "--off",
    help="Take PROFILE's library off from the Memory of the World Library.",
    is_flag=True,
)
def submit(profile, librarian, off):
    """Submit all the local (metadata) changes for PROFILE's Calibre collection to the Memory of the World Library (https://library.memoryoftheworld.org)"""
    run_submit(APP_NAME, profile, librarian, off)


@cli.command(
    short_help="Release local Calibre collection to the Memory of the World Library. For more help, type: accorder release --help"
)
@click.argument("profile")
@click.option(
    "--directory",
    "-d",
    help="Add/change PROFILE's directory path to be uploaded to the server. It is automatically saved into configuration file for the future use.",
    type=click.Path(exists=True, resolve_path=True, file_okay=False),
)
@click.option(
    "--librarian",
    "-l",
    help="Add/change librarian's name (max. 42 chars) for the given PROFILE. Once provided, librarian's name is automatically saved into configuration file for the future use.",
)
@click.option(
    "--delete-residue",
    help="Delete any remote directories/files which are not present locally anymore.",
    is_flag=True,
)
@click.option(
    "--off",
    help="Take PROFILE's library off from the Memory of the World Library.",
    is_flag=True,
)
def release(profile, directory, librarian, delete_residue, off):
    """Release PROFILE's Calibre collection to the Memory of the World Library. First, it builds locally the standalone portable webapp (as if `accorder build PROFILE` was run), then it uploads all the files to the server (`accorder upload PROFILE`) and finally it submits all the metadata to the Memory of the World Library (`accorder submit PROFILE`)."""
    run_build(APP_NAME, profile, directory, librarian, jsonpath=None, with_bibtex=None)
    run_upload(APP_NAME, profile, directory, delete_residue)
    run_submit(APP_NAME, profile, librarian, off)


if __name__ == "__main__":
    cli()



import subprocess
from threading import Thread
import time
import os
import socket
from dataclasses import dataclass
import click
from accorder.utils.configs import edit_config
from accorder.utils.configs import check_sections


@dataclass(init=False)
class RsyncTunneled:
    tunnel_port_forward: str
    tunnel_account: str
    tunnel_ssh_port: str
    rsync_password: str
    rsync_local_directory: str
    rsync_user: str
    rsync_endpoint: str
    start_time: str
    rsync_delete: str
    keep_thread: bool = True
    keep: bool = True
    timeout: int = 10

    def run_ssh(self):
        def _setup_timeout(ssh_process):
            self.start_time = time.time()
            while self.keep_thread:
                delta = "*" * (self.timeout - round(time.time() - self.start_time))
                click.echo(
                    "\rConnecting: [{}] ".format(delta.ljust(self.timeout, " ")),
                    nl=False,
                )
                if round(time.time() - self.start_time, 3) > self.timeout:
                    self.keep_thread = False
                    self.keep = False
                    ssh_process.terminate()

        self.local_port, self.server_ip, self.remote_port = self.tunnel_port_forward.split(
            ":"
        )
        for i in range(32):
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                self.local_port = int(self.local_port)
                if sock.connect_ex(('localhost', self.local_port)):
                    self.tunnel_port_forward = f"{self.local_port}:{self.server_ip}:{self.remote_port}"
                    break
                else:
                    self.local_port += 1
        else:
            click.echo(
                "\rLocal ports 8873-8905 are already in use. Close the process to release one of the ports and then try again."
            )
            return False

        self.ssh_process = subprocess.Popen(
            [
                "ssh",
                "-v",
                "-N",
                "-o",
                "ExitOnForwardFailure=yes",
                "-o",
                "ServerAliveINterval=60",
                "-o",
                "TCPKeepAlive=yes",
                "-o",
                "StrictHostKeyChecking=no",
                "-o",
                "UserKnownHostsFile=/dev/null",
                "-L",
                self.tunnel_port_forward,
                self.tunnel_account,
                "-p",
                self.tunnel_ssh_port,
            ],
            stderr=subprocess.PIPE,
            universal_newlines=True,
            bufsize=0,
        )

        counter = Thread(target=_setup_timeout, args=(self.ssh_process,))
        counter.start()

        line = ""
        while self.keep:
            rbytes = self.ssh_process.stderr.read(1)
            line += rbytes
            if rbytes == "\n":
                if "Entering interactive session." in line.strip():
                    self.keep_thread = False
                    click.echo("\rSecure connection established.")
                    self.keep = False
                    break
                elif "Address already in use" in line.strip():
                    self.keep_thread = False
                    click.echo(
                        f"\rLocal port {self.local_port} is already in use. Close the process to release the port and then try again."
                    )
                    self.keep = False
                    return False

                line = ""

        time.sleep(1)
        if self.ssh_process.poll() is not None:
            click.echo("\rSSH tunnel failed. Check your internet connection.")
            return False

        return True

    def run_rsync(self):
        click.echo(
            f"\nUploading local directory:\n{self.rsync_local_directory}\n"
        )
        os.environ["RSYNC_PASSWORD"] = self.rsync_password
        command = f'rsync -zvrith --omit-dir-times --progress {self.rsync_delete} --no-perms --inplace --no-whole-file "{self.rsync_local_directory}" rsync://{self.rsync_user}@localhost:{self.local_port}/{self.rsync_endpoint}/'
        ec = os.system(command)
        self.ssh_process.terminate()
        if ec == 0:
            click.echo(f"Upload succeeded.")
        else:
            click.echo(f"Upload failed! Please, try again.")
            exit(1)
        click.echo("")


@dataclass(init=False)
class Rclone:
    rclone_configuration: str
    rclone_profile: str
    rclone_url: str
    rclone_local_directory: str
    rclone_copy_or_sync: str

    def run_rclone(self):
        click.echo(
            f"\nLocal: {self.rclone_local_directory}\nRemote: {self.rclone_url}\n"
        )
        command = f"rclone --config={self.rclone_configuration} {self.rclone_copy_or_sync} --fast-list --progress --checkers 64 {self.rclone_local_directory} {self.rclone_profile}_rclone:"
        os.system(command)
        click.echo("")


def run_upload(app_name, profile, directory, delete_residue):
    config = edit_config(app_name, profile, directory=directory)
    sections = check_sections(config, profile)

    if "rsync" not in sections and "rclone" not in sections:
        click.echo(
            f"{profile} doesn't have any configuration set for the upload. Check out: `accorder configuration --help`"
        )
        click.get_current_context().exit()

    motw_message = ""
    if "motw" in sections:
        motw_subdomain = config[profile]["motw"]["subdomain"]
        motw_domain = config[profile]["motw"]["domain"]
        motw_message = f"TO https://{motw_subdomain}.{motw_domain}"

    click.echo(f">>>> UPLOAD FILES {motw_message}")
    if config[profile]["meta"]["upload"] == "rsync":
        click.echo(">>>> VIA RSYNC")
        r = RsyncTunneled()
        r.tunnel_port_forward = config[profile]["tunnel"]["port_forward"]
        r.tunnel_account = config[profile]["tunnel"]["account"]
        # r.tunnel_account = "tunnel@1.1.1.1"
        r.tunnel_ssh_port = config[profile]["tunnel"]["ssh_port"]

        r.rsync_password = config[profile]["rsync"]["password"]
        r.rsync_local_directory = config[profile]["calibre"]["local_directory"]
        r.rsync_user = config[profile]["rsync"]["user"]
        r.rsync_endpoint = config[profile]["rsync"]["endpoint"]
        r.rsync_delete = ""
        if delete_residue:
            r.rsync_delete = "--delete"

        ssh = r.run_ssh()
        if ssh:
            r.run_rsync()
    elif config[profile]["meta"]["upload"] == "rclone":
        click.echo(">>>> VIA RCLONE")
        r = Rclone()
        r.rclone_profile = profile
        r.rclone_url = config[profile]["rclone"]["url"]
        r.rclone_local_directory = config[profile]["calibre"]["local_directory"]
        r.rclone_configuration = os.path.join(
            click.get_app_dir(app_name), "accorder.ini"
        )
        r.rclone_copy_or_sync = "copy"
        if delete_residue:
            r.rclone_copy_or_sync = "sync"
        r.run_rclone()



# -*- coding: utf-8 -*-

import hmac
import hashlib
import os
import sqlite3
import time
import uuid
import zlib

import urllib.request
import urllib.parse


def calibre_to_json(library_uuid, library_secret, librarian, db_path):
    bid = """
    select * from books
    """
    bid_comments = """
    select books.id as book_id,
           comments.text as comments
    from books
    join comments on books.id = comments.book
    """
    bid_tag = """
    select books.id as book_id,
           tags.name as tag
    from books
    join (tags join books_tags_link on tags.id = books_tags_link.tag)
    books_tags_link on books.id = books_tags_link.book
    """
    bid_identifiers = """
    select books.id as book_id,
           identifiers.type as id_type,
           identifiers.val as id_val
    from books
    join identifiers on identifiers.book = books.id
    """
    bid_formats = """
    select books.id as book_id,
           data.format as file_format,
           data.uncompressed_size as file_size,
           data.name as file_name
    from books
    join data on data.book = books.id
    """
    bid_publishers = """
    select books.id as book_id,
           publishers.name
    from books
    join (publishers join books_publishers_link on publishers.id = books_publishers_link.publisher)
    books_publishers_link on books.id = books_publishers_link.book
    """
    bid_authors = """
    select books.id as book_id,
           authors.name
    from books join (authors join books_authors_link on authors.id = books_authors_link.author)
    books_authors_link on books.id = books_authors_link.book
    """
    bid_series = """
    select books.id as book_id,
           series.name
    from books join (series join books_series_link on series.id = books_series_link.series)
    books_series_link on books.id = books_series_link.book
    """
    bid_languages = """
    select books.id as book_id,
           languages.lang_code as lang_code
    from books join (languages join books_languages_link on languages.id = books_languages_link.lang_code)
    books_languages_link on books.id = books_languages_link.book
    """

    conn = sqlite3.connect(os.path.join(db_path))
    cur = conn.cursor()
    sql_books = (book for book in cur.execute(bid))

    books = {}
    book_dict = [
        "application_id",
        "title",
        "title_sort",
        "timestamp",
        "pubdate",
        "series_index",
        "author_sort",
        "isbn",
        "iccn",
        "path",
        "flags",
        "uuid",
        "has_cover",
        "last_modified",  # `select * from books` ends here
        "library_uuid",
        "librarian",
        "_id",
        "tags",
        "abstract",
        "publisher",
        "authors",
        # 'card',
        "formats",
        "cover_url",
        "identifiers",
        "languages",
    ]

    _ = [
        books.update(
            {
                book[0]: dict(
                    zip(
                        book_dict,
                        book
                        + (
                            library_uuid,  # library_uuid
                            librarian,  # librarian
                            str(
                                uuid.UUID(
                                    hmac.new(
                                        # library_secret.encode(), book[11].encode()
                                        library_uuid.encode(),
                                        book[11].encode(),
                                        digestmod=hashlib.md5
                                    ).hexdigest(),
                                    version=4,
                                )
                            ),
                            [],  # tags
                            "",  # comments
                            "",  # publisher
                            [],  # authors
                            # {},  # card
                            [],  # formats
                            "{}/cover.jpg".format(book[9]),  # cover_url
                            [],  # identifiers
                            [],  # languages
                        ),
                    )
                )
            }
        )
        for book in sql_books
    ]

    sql_tags = [tag for tag in cur.execute(bid_tag)]
    [books[tag[0]]["tags"].append(tag[1][:100]) for tag in sql_tags]

    # bleach.sanitizer.ALLOWED_TAGS:
    # ['a', 'abbr', 'acronym', 'b', 'blockquote', 'code', 'em', 'i',
    #  'li', 'ol', 'strong', 'ul']
    # allowed_tags = bleach.sanitizer.ALLOWED_TAGS + ['p', 'div', 'br', 'pre']
    sql_comments = (comment for comment in cur.execute(bid_comments))
    [
        books[comment[0]].update(
            # {
            #     'comments': bleach.clean(comment[1][:10000],
            #                              strip=True,
            #                              tags=allowed_tags)
            # })
            {"abstract": comment[1][:10000]}
        )
        for comment in sql_comments
    ]
    sql_publishers = (publisher for publisher in cur.execute(bid_publishers))
    [
        books[publisher[0]].update({"publisher": publisher[1][:100]})
        for publisher in sql_publishers
    ]

    sql_series = (series for series in cur.execute(bid_series))
    [books[series[0]].update({"series": series[1][:100]}) for series in sql_series]

    sql_authors = (author for author in cur.execute(bid_authors))
    [books[author[0]]["authors"].append(author[1][:200]) for author in sql_authors]

    sql_identifiers = (identifier for identifier in cur.execute(bid_identifiers))
    [
        books[identifier[0]]["identifiers"].append(
            {"scheme": identifier[1][:100], "code": identifier[2][:1000]}
        )
        for identifier in sql_identifiers
    ]

    sql_formats = (frmat for frmat in cur.execute(bid_formats))
    [
        books[frmat[0]]["formats"].append(
            {
                "format": "{}".format(frmat[1].lower()),
                "file_name": "{}.{}".format(frmat[3], frmat[1].lower()),
                "dir_path": "{}/".format(books[frmat[0]]["path"]),
                "size": frmat[2],
            }
        )
        for frmat in sql_formats
    ]

    sql_languages = (language for language in cur.execute(bid_languages))
    [books[language[0]]["languages"].append(language[1]) for language in sql_languages]

    books_list = []
    remove_keys = [
        "application_id",
        "isbn",
        "iccn",
        "path",
        "flags",
        "has_cover",
        "uuid",
        "author_sort",
        "timestamp",
        "series_index",
    ]
    # modify_keys = ['timestamp', 'pubdate', 'last_modified']
    for book in list(books.values()):
        if "private" in [x for x in map(str.lower, book["tags"])]:
            continue
        for k in remove_keys:
            book.pop(k, None)
        books_list.append(book)

    return books_list


def add_library(host_api, library_uuid, library_secret):
    headers = {"Library-Secret": library_secret}
    url = f"{host_api}/library/add/{library_uuid}"
    req = urllib.request.Request(url, headers=headers)
    res = urllib.request.urlopen(req)
    return (res.getcode(), res.read())


def remove_library(host_api, library_uuid, library_secret):
    headers = {"Library-Secret": library_secret}
    url = f"{host_api}/library/remove/{library_uuid}"
    req = urllib.request.Request(url, headers=headers)
    res = urllib.request.urlopen(req)
    return (res.getcode(), res.read())


def library_on(host_api, library_uuid, library_secret, librarian="", url="", reset=""):
    headers = {"Library-Secret": library_secret}
    url_args = {"librarian": librarian, "url": url, "reset": reset}
    query_string = urllib.parse.urlencode(url_args)
    url = f"{host_api}/library/on/{library_uuid}?{query_string}"
    req = urllib.request.Request(url, headers=headers)
    res = urllib.request.urlopen(req)
    return (res.getcode(), res.read())


def library_off(host_api, library_uuid, library_secret):
    headers = {"Library-Secret": library_secret}
    url = f"{host_api}/library/off/{library_uuid}"
    req = urllib.request.Request(url, headers=headers)
    res = urllib.request.urlopen(req)
    return (res.getcode(), res.read())


def bookids(host_api, library_uuid, library_secret):
    headers = {"Library-Secret": library_secret}
    url = f"{host_api}/library/bookids/{library_uuid}"
    req = urllib.request.Request(url, headers=headers)
    res = urllib.request.urlopen(req)
    return (res.getcode(), res.read())


def load_books(host_api, verb, library_uuid, library_secret):
    headers = {"Library-Secret": library_secret}
    headers.update({"Library-Encoding": "zlib"})
    url = f"{host_api}/books/load/{verb}/{library_uuid}"
    req = urllib.request.Request(url, headers=headers)
    res = urllib.request.urlopen(req)
    return (res.getcode(), res.read())


def upload_books(host_api, library_uuid, library_secret, payload, zipit=False):
    headers = {"Library-Secret": library_secret}
    if zipit:
        # t = time.time()
        headers.update({"Library-Encoding": "zlib"})
        payload = zlib.compress(payload.encode("utf-8"))
        # print("zlib payload in {} seconds.".format(round(time.time() - t, 3)))
    else:
        payload = payload.encode()

    url = f"{host_api}/upload/{library_uuid}"
    req = urllib.request.Request(url, headers=headers, data=payload)
    res = urllib.request.urlopen(req)
    return (res.getcode(), res.read())


def add_books(host_api, library_uuid, library_secret, payload, zipit=False):
    headers = {"Library-Secret": library_secret}
    if zipit:
        t = time.time()
        headers.update({"Library-Encoding": "zlib"})
        payload = zlib.compress(payload.encode("utf-8"))
        print("zlib payload in {} seconds.".format(round(time.time() - t, 3)))
    else:
        payload = payload.encode()

    url = f"{host_api}/add/{library_uuid}"
    req = urllib.request.Request(url, headers=headers, data=payload)
    res = urllib.request.urlopen(req)
    return (res.getcode(), res.read())


def remove_books(host_api, library_uuid, library_secret, payload, zipit=False):
    headers = {"Library-Secret": library_secret}
    if zipit:
        t = time.time()
        headers.update({"Library-Encoding": "zlib"})
        payload = zlib.compress(payload.encode("utf-8"))
        print("zlib payload in {} seconds.".format(round(time.time() - t, 3)))
    else:
        payload = payload.encode()

    url = f"{host_api}/books/remove/{library_uuid}"
    req = urllib.request.Request(url, headers=headers, data=payload)
    res = urllib.request.urlopen(req)
    return (res.getcode(), res.read())



import json
import click
from dataclasses import dataclass
from accorder.utils.motw_client import add_library
from accorder.utils.motw_client import bookids
from accorder.utils.motw_client import remove_books
from accorder.utils.motw_client import library_on
from accorder.utils.motw_client import library_off
from accorder.utils.motw_client import upload_books
from accorder.utils.motw_client import load_books
from accorder.utils.configs import edit_config
from accorder.utils.configs import check_profile


@dataclass(init=False)
class Submit:
    submit_off: bool = False
    submit_domain: str
    submit_subdomain: str
    submit_api: str
    submit_upload_api: str
    submit_librarian: str
    submit_library_uuid: str
    submit_library_secret: str
    submit_data_js: str

    def submit(self):
        lsb_url = f"//{self.submit_subdomain}.{self.submit_domain}/"
        click.echo(f">>>  {self.submit_librarian} at https:{lsb_url}")

        try:
            add = add_library(
                self.submit_api, self.submit_library_uuid, self.submit_library_secret
            )
            if add[0] == 200:
                click.echo(
                    f"{self.submit_librarian} from {lsb_url} added to {self.submit_api}"
                )
            elif add[0] == 202:
                # click.echo(add[1].decode())
                pass
        except Exception as e:
            click.echo(f"add_library failed: {e}")

        if self.submit_off:
            lib_res = library_off(
                self.submit_api, self.submit_library_uuid, self.submit_library_secret
            )
            if lib_res[0] == 200:
                click.echo(
                    f"{self.submit_librarian}'s library at {self.submit_api} is now OFF."
                )
            elif lib_res[0] == 202:
                click.echo(
                    f"{self.submit_librarian}'s library at {self.submit_api} has already been OFF. Don't flog a dead horse!"
                )
            click.get_current_context().exit()

        bids = bookids(
            self.submit_api, self.submit_library_uuid, self.submit_library_secret
        )
        if bids[0] == 200:
            bids = json.loads(bids[1].decode())
        else:
            raise Exception(
                f"Request to {self.submit_api} for book ids of {self.submit_library_uuid} failed."
            )
        # books_list = calibre_to_json(library_uuid, library_secret, librarian, db_path)
        books_list = []
        for i in range(1, 9):
            with open(f"{self.submit_data_js}data{i}.js") as f:
                f.seek(15)
                books_list += json.loads(f.read())["books"]
        rids = set(
            [
                s.split("___")[0]
                for s in list(
                    set(bids)
                    - set(
                        [
                            "{}___{}".format(b["_id"], b["last_modified"])
                            for b in books_list
                        ]
                    )
                )
            ]
        )
        if len(rids) > 0:
            click.echo(f"{len(rids)} books changed or deleted.")
        else:
            click.echo("No books changed or deleted.")
        if rids:
            reset = remove_books(
                self.submit_api,
                self.submit_library_uuid,
                self.submit_library_secret,
                json.dumps(list(rids)),
                True,
            )
            if reset[0] == 200:
                reset = json.loads(reset[1].decode())
            else:
                raise Exception(
                    f"Request to {self.submit_api} to remove books from {self.submit_library_uuid} failed."
                )

            if reset["reset"]:
                # library_off(library_uuid, library_secret)
                lib_res = library_on(
                    self.submit_api,
                    self.submit_library_uuid,
                    self.submit_library_secret,
                    self.submit_librarian,
                    lsb_url,
                    "on",
                )
                if lib_res[0] == 200:
                    click.echo(
                        f"{self.submit_librarian}'s {len(rids)} books updated or deleted from {self.submit_api}"
                    )
            else:
                lib_res = library_on(
                    self.submit_api,
                    self.submit_library_uuid,
                    self.submit_library_secret,
                    self.submit_librarian,
                    lsb_url,
                )
                if lib_res[0] == 202:
                    click.echo(
                        f"{self.submit_librarian}'s {len(rids)} books updated or deleted from {self.submit_api}"
                    )

        bids = bookids(
            self.submit_api, self.submit_library_uuid, self.submit_library_secret
        )
        if bids[0] == 200:
            bids = json.loads(bids[1].decode())
        else:
            raise Exception(
                f"Request to {self.submit_api} for book ids of {self.submit_library_uuid} failed."
            )
        sids = set(
            [
                s.split("___")[0]
                for s in list(
                    set(
                        [
                            "{}___{}".format(b["_id"], b["last_modified"])
                            for b in books_list
                        ]
                    )
                    - set(bids)
                )
            ]
        )

        if len(sids) > 0:
            click.echo(f"{len(sids)} books to upload... ", nl=False)
        else:
            click.echo("No books to upload.")

        if len(sids) == 0 and len(rids) == 0:
            lib_res = library_on(
                self.submit_api,
                self.submit_library_uuid,
                self.submit_library_secret,
            )

            if lib_res[0] == 200:
                click.echo(f"{self.submit_librarian}'s library was OFF. Now it's back online. YAY!")

        if sids:
            # reset = add_books(
            upload = upload_books(
                self.submit_upload_api,
                self.submit_library_uuid,
                self.submit_library_secret,
                json.dumps([b for b in books_list if b["_id"] in sids]),
                True,
            )
            if upload[0] == 200:
                reset = load_books(
                    self.submit_api,
                    "add",
                    self.submit_library_uuid,
                    self.submit_library_secret,
                )
            if reset[0] == 200:
                reset = json.loads(reset[1].decode())
                click.echo("Uploaded.")
            else:
                raise Exception(
                    f"Request to load books of just uploaded {self.submit_library_uuid} failed."
                )

            if reset["reset"]:
                # library_off(library_uuid, library_secret)
                lib_res = library_on(
                    self.submit_api,
                    self.submit_library_uuid,
                    self.submit_library_secret,
                    self.submit_librarian,
                    lsb_url,
                    "on",
                )
                if lib_res[0] == 200:
                    click.echo(
                        f"{self.submit_librarian}'s {len(sids)} books added to {self.submit_api}."
                    )
            else:
                lib_res = library_on(
                    self.submit_api,
                    self.submit_library_uuid,
                    self.submit_library_secret,
                    self.submit_librarian,
                    lsb_url,
                )
                if lib_res[0] == 200:
                    click.echo(
                        f"{self.submit_librarian}'s {len(sids)} books added to {self.submit_api}."
                    )

        click.echo("")


def run_submit(app_name, profile, librarian, off):
    config = edit_config(app_name, profile, librarian)
    check_profile(config, profile)

    r = Submit()
    if off:
        r.submit_off = off
    r.submit_api = config[profile]["motw"]["api"]
    r.submit_upload_api = config[profile]["motw"]["upload_api"]
    r.submit_subdomain = config[profile]["motw"]["subdomain"]
    r.submit_domain = config[profile]["motw"]["domain"]
    r.submit_librarian = config[profile]["calibre"]["librarian"]
    r.submit_library_uuid = config[profile]["calibre"]["library_uuid"]
    r.submit_library_secret = config[profile]["calibre"]["library_secret"]
    r.submit_data_js = f'{config[profile]["calibre"]["local_directory"]}/static/'

    click.echo(f">>>> SUBMIT METADATA TO {r.submit_api}")
    r.submit()
